import xmpp

from moodle_dl.utils import check_verbose


class XmppShooter:
    def __init__(self, jid, password, recipient):
        self.jid = xmpp.protocol.JID(jid)
        self.connection = xmpp.Client(server=self.jid.getDomain(), debug=check_verbose())
        self.is_connected = False
        self.password = password
        self.recipient = xmpp.protocol.JID(recipient)

    def send(self, message):
        if not self.is_connected:
            self.connection.connect()
            self.connection.auth(user=self.jid.getNode(), password=self.password, resource=self.jid.getResource())
            if not hasattr(self.connection, 'Bind') or getattr(self.connection, 'Bind').session != 1:
                raise ConnectionError('XMPP Session could not be opend')
            self.is_connected = True

        self.connection.send(xmpp.protocol.Message(to=self.recipient, body=message))



import re
from urllib.parse import urljoin, urlparse

from yt_dlp.extractor.common import InfoExtractor
from yt_dlp.utils import ExtractorError, determine_ext, url_or_none, urlencode_postdata


class OwnCloudIE(InfoExtractor):
    _INSTANCES_RE = '|'.join(
        (
            r'(?:[^\.]+\.)?sciebo\.de',
            r'cloud\.uni-koblenz-landau\.de',
        )
    )
    _VALID_URL = rf'https?://(?:{_INSTANCES_RE})/s/(?P<id>[\w.-]+)'

    _TESTS = [
        {
            'url': 'https://ruhr-uni-bochum.sciebo.de/s/wWhqZzh9jTumVFN',
            'info_dict': {
                'id': 'wWhqZzh9jTumVFN',
                'ext': 'mp4',
                'title': 'CmvpJST.mp4',
            },
        },
        {
            'url': 'https://ruhr-uni-bochum.sciebo.de/s/WNDuFu0XuFtmm3f',
            'info_dict': {
                'id': 'WNDuFu0XuFtmm3f',
                'ext': 'mp4',
                'title': 'CmvpJST.mp4',
            },
            'params': {
                'videopassword': '12345',
            },
        },
    ]

    def _real_extract(self, url):
        video_id = self._match_id(url)
        webpage, urlh = self._download_webpage_handle(url, video_id)

        if re.search(r'<label[^>]+for="password"', webpage):
            webpage = self._verify_video_password(webpage, urlh.geturl(), video_id)

        hidden_inputs = self._hidden_inputs(webpage)
        title = hidden_inputs.get('filename')

        return {
            'id': video_id,
            'title': title,
            'url': (
                url_or_none(
                    hidden_inputs.get('downloadURL')
                    or urlparse(url)._replace(path=urljoin(urlparse(url).path, 'download')).geturl()
                )
            ),
            'ext': determine_ext(title),
        }

    def _verify_video_password(self, webpage, url, video_id):
        password = self.get_param('videopassword')
        if password is None:
            raise ExtractorError(
                'This video is protected by a password, use the --video-password option', expected=True
            )

        validation_response = self._download_webpage(
            url,
            video_id,
            'Validating Password',
            'Wrong password?',
            data=urlencode_postdata(
                {
                    'requesttoken': self._hidden_inputs(webpage)['requesttoken'],
                    'password': password,
                }
            ),
        )

        if re.search(r'<label[^>]+for="password"', validation_response):
            warning = self._search_regex(
                r'<div[^>]+class="warning">([^<]*)</div>',
                validation_response,
                'warning',
                default='The password is wrong',
            )
            raise ExtractorError(f'Opening the video failed, {self.IE_NAME} said: {warning!r}', expected=True)
        return validation_response



import urllib

from requests.exceptions import RequestException

from moodle_dl.utils import SslHelper


class TelegramShooter:
    RQ_HEADER = {
        'User-Agent': (
            'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/78.0.3904.108 Safari/537.36'
        ),
        'Content-Type': 'application/x-www-form-urlencoded',
    }

    def __init__(self, telegram_token: str, telegram_chatid: str):
        self.telegram_token = telegram_token
        self.telegram_chatid = telegram_chatid

    def send(self, message: str):
        payload = {'chat_id': self.telegram_chatid, 'text': message, 'parse_mode': 'HTML'}

        url = f'https://api.telegram.org/bot{self.telegram_token}/sendMessage'
        data_urlencoded = urllib.parse.urlencode(payload)

        session = SslHelper.custom_requests_session(skip_cert_verify=False, allow_insecure_ssl=False)
        try:
            response = session.post(url, data=data_urlencoded, headers=self.RQ_HEADER, timeout=60)
        except RequestException as error:
            raise ConnectionError(f"Connection error: {str(error)}") from None

        self._check_errors(response)

    @staticmethod
    def _check_response_code(response):
        # Normally Telegram answer with response 200
        if response.status_code not in [200, 400]:
            raise RuntimeError(
                'An Unexpected Error happened on side of the Telegram System!'
                + f' Status-Code: {str(response.status_code)}'
                + f'\nHeader: {response.headers}'
                + f'\nResponse: {response.text}'
            )

    def _check_errors(self, response) -> object:
        """
        The first time parsing the result of a REST request.
        It is checked for known errors.
        @param response: The JSON response of the Moodle system
        @return: The parsed JSON object
        """

        self._check_response_code(response)

        # Try to parse the JSON
        try:
            response_extracted = response.json()
        except Exception as error:
            raise RuntimeError(
                'An Unexpected Error occurred while trying'
                + ' to parse the json response! Telegram'
                + f' response: {response.read()}.\nError: {error}'
            )
        # Check for known errors
        if "ok" in response_extracted:
            ok = response_extracted.get("ok", False)

            if not ok:
                raise RequestRejectedError(
                    'The Telegram System rejected the Request.'
                    + f" Details: {response_extracted.get('description', 'None')}"
                )

        return response_extracted


class RequestRejectedError(Exception):
    """An Exception which gets thrown if the Telegram-System answered with an
    Error to our Request"""

    pass



from yt_dlp.extractor.common import InfoExtractor
from yt_dlp.utils import determine_ext, int_or_none, url_or_none


class SharePointFilesIE(InfoExtractor):
    _VALID_URL = r'https?://[^\.]+\.sharepoint\.com/:[pwx]:/g/[^/]+/[^/]+/(?P<id>[^/?]+)'

    _TESTS = [
        {
            'url': 'https://lut-my.sharepoint.com/:p:/g/personal/juha_eerola_student_lab_fi/'
            + 'EdHr6d6EV89LlfipfMzVsasBymdnQBpUVor9LrhiX24Z8w?e=003Xgv',
            'info_dict': {
                'id': 'EdHr6d6EV89LlfipfMzVsasBymdnQBpUVor9LrhiX24Z8w',
                'ext': 'pptx',
                'title': '2. K23 Kivun luokittelu keston mukaan.pptx',
                'modified_timestamp': 1675785393000,
            },
        },
        {
            'url': 'https://lut-my.sharepoint.com/:w:/g/personal/juha_eerola_student_lab_fi/'
            + 'EfjpAV2zXSpKmMflo08J15QBKWpb-0fFSsx-Q4yvmzc1mw?e=n14Mto',
            'info_dict': {
                'id': 'EfjpAV2zXSpKmMflo08J15QBKWpb-0fFSsx-Q4yvmzc1mw',
                'ext': 'docx',
                'title': 'LABin_opinnäytetyöraportin_kirjoituspohja_260822.docx',
                'modified_timestamp': 1675786271000,
            },
        },
        {
            'url': 'https://lut-my.sharepoint.com/:x:/g/personal/juha_eerola_student_lab_fi/'
            + 'EeMmrRu3BptJoH5WPTsdTyEB460rKPzMu-GmHIH2xcJGXQ?e=tVbK9U',
            'info_dict': {
                'id': 'EeMmrRu3BptJoH5WPTsdTyEB460rKPzMu-GmHIH2xcJGXQ',
                'ext': 'xlsx',
                'title': 'Kalenteri ryhmätöitä varten - Imatra.xlsx',
                'modified_timestamp': 1665724530000,
            },
            'params': {'skip_download': 'file small'},
        },
    ]

    def _real_extract(self, url):
        video_id = self._match_valid_url(url).group('id')
        webpage = self._download_webpage(url, video_id)

        metadata = self._search_json(r'_wopiContextJson\s*=', webpage, 'metadata', video_id)
        download_url = url_or_none(metadata.get('FileGetUrl'))
        return {
            'id': video_id,
            'title': metadata.get('FileName'),
            'modified_timestamp': metadata.get('LastModified'),
            'formats': [
                {
                    'url': download_url,
                    'ext': determine_ext(metadata.get('FileName')),
                    'filesize': int_or_none(metadata.get('FileSize')),
                }
            ],
        }



import asyncio
from typing import Dict, List

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.mods.common import MoodleMod
from moodle_dl.moodle.request_helper import RequestHelper
from moodle_dl.types import Course

from moodle_dl.moodle.mods.assign import AssignMod  # noqa: F401 isort:skip
from moodle_dl.moodle.mods.data import DataMod  # noqa: F401 isort:skip
from moodle_dl.moodle.mods.folder import FolderMod  # noqa: F401 isort:skip
from moodle_dl.moodle.mods.forum import ForumMod  # noqa: F401 isort:skip
from moodle_dl.moodle.mods.lesson import LessonMod  # noqa: F401 isort:skip
from moodle_dl.moodle.mods.page import PageMod  # noqa: F401 isort:skip
from moodle_dl.moodle.mods.quiz import QuizMod  # noqa: F401 isort:skip
from moodle_dl.moodle.mods.workshop import WorkshopMod  # noqa: F401 isort:skip

ALL_MODS = [Class for name, Class in globals().items() if name.endswith('Mod') and name != 'MoodleMod']


def get_all_mods_classes() -> List[MoodleMod]:
    return ALL_MODS


def get_all_mods(
    request_helper: RequestHelper,
    moodle_version: int,
    user_id: int,
    last_timestamps: Dict[str, Dict[int, int]],
    config: ConfigHelper,
) -> List[MoodleMod]:
    result = []
    for mod in ALL_MODS:
        result.append(mod(request_helper, moodle_version, user_id, last_timestamps, config))
    return result


async def fetch_mods_files(mods_to_fetch: List[MoodleMod], courses_to_load: List[Course]) -> Dict[str, Dict]:
    "@return: Dictionary of all fetched files, indexed by mod name, then by courses, then module id"
    mods_results = await asyncio.gather(*[mod.fetch_mod_entries(courses_to_load) for mod in mods_to_fetch])
    result = {}
    for idx, mod in enumerate(mods_to_fetch):
        result[mod.MOD_NAME] = mods_results[idx]
    return result


def get_mod_plurals():
    result = {}
    for mod in ALL_MODS:
        result[mod.MOD_NAME] = mod.MOD_PLURAL_NAME
    return result



import hashlib
import html
import logging
import mimetypes
import re
import urllib.parse as urlparse
from typing import Dict, List

from moodle_dl.types import Course, File, MoodleURL


class ResultBuilder:
    """
    Combines all fetched mod files and core course files to one result based on File objects
    """

    def __init__(self, moodle_url: MoodleURL, version: int, mod_plurals: Dict):
        self.version = version
        self.moodle_url = moodle_url
        self.moodle_domain = moodle_url.domain
        self.mod_plurals = mod_plurals

    def get_files_in_sections(self, course_sections: List[Dict], fetched_mods: Dict[str, Dict]) -> List[File]:
        """
        Iterates over all sections of a course to find files (or modules).
        @param course_sections: Contains the sections of the course
        @param fetched_mods: Contains the fetched mods of the course
        @return: A list of files of the course.
        """
        files = []
        for section in course_sections:
            location = {
                'section_id': section.get('id', 0),
                'section_name': section.get('name', ''),
            }
            section_modules = section.get('modules', [])
            files += self._get_files_in_modules(section_modules, fetched_mods, **location)

            section_summary = section.get('summary', '')
            if section_summary is not None and section_summary != '':
                location.update(
                    {
                        'module_id': 0,
                        'module_name': 'Section summary',
                        'module_modname': 'section_summary',
                    }
                )
                files += self._handle_description(section_summary, **location)

        files += self._get_files_not_on_main_page(fetched_mods)

        return files

    def _get_files_in_modules(self, section_modules: List, fetched_mods: Dict[str, Dict], **location) -> List[File]:
        """
        Iterates over all modules to find files (or content) in them.
        @param section_modules: The modules of the section.
        @param fetched_mods: Contains the fetched mods of the course
        @param location: contains
            section_id: int,
            section_name: str,
        @return: A list of files of the section.
        """
        files = []
        for module in section_modules:
            location['module_id'] = module.get('id', 0)
            location['module_name'] = module.get('name', '')
            location['module_modname'] = module.get('modname', '')

            module_url = module.get('url', '')
            module_contents = module.get('contents', [])
            module_description = module.get('description', None)

            # handle not supported modules that results in an index.html special
            if location['module_modname'] in ['moodecvideo']:
                location['module_modname'] = 'index_mod-' + location['module_modname']

            if location['module_modname'] in ['page'] and self.version < 2017051500:
                # legacy pages
                location['module_modname'] = 'index_mod-' + location['module_modname']

            if module_description is not None and location['module_modname'] not in fetched_mods:
                # Handle descriptions of Files, Labels and all mods that we do not handle in separately
                files += self._handle_description(module_description, **location)

            if location['module_modname'] in ['kalvidres', 'helixmedia', 'lti']:
                location['module_modname'] = 'cookie_mod-' + location['module_modname']
                files += self._handle_cookie_mod(module_url, **location)

            elif location['module_modname'].startswith(('resource', 'akarifolder', 'url', 'index_mod')):
                files += self._handle_files(module_contents, **location)

            elif location['module_modname'].startswith(('folder')):
                # Modules whose content is directly linked, and additional content exists in their module.
                if location['module_modname'] in fetched_mods:
                    # find mod module with same module_id
                    mod = fetched_mods.get(location['module_modname'], {}).get(location['module_id'], {})
                    mod['on_main_page'] = True
                    mod_files = mod.get('files', [])
                    module_contents += mod_files

                files += self._handle_files(module_contents, **location)

            elif location['module_modname'] in fetched_mods:
                # find mod module with same module_id
                mod = fetched_mods.get(location['module_modname'], {}).get(location['module_id'], {})
                mod['on_main_page'] = True
                mod_files = mod.get('files', [])
                files += self._handle_files(mod_files, **location)
            else:
                if location['module_modname'] not in ['label']:
                    logging.debug(
                        'Got unhandled module: name=%s mod=%s url=%s',
                        location['module_name'],
                        location['module_modname'],
                        module_url,
                    )

        return files

    def get_mod_plural_name(self, mod_name: str) -> str:
        if mod_name in self.mod_plurals:
            return self.mod_plurals[mod_name].capitalize()
        return mod_name.capitalize()

    def _get_files_not_on_main_page(self, fetched_mods: Dict[str, Dict]) -> List[File]:
        """
        Iterates over all mods to find files (or content) that are not listed on the main page.
        @param fetched_mods: Contains the fetched mods of the course
        @return: A list of files of mod modules not on the main page.
        """
        files = []
        for mod_name, mod_modules in fetched_mods.items():
            location = {
                'section_id': -1,
                'section_name': f"{self.get_mod_plural_name(mod_name)} not on main page",
            }

            for _, module in mod_modules.items():
                if 'on_main_page' in module:
                    continue
                location.update(
                    {
                        'module_id': module.get('id', 0),
                        'module_name': module.get('name', ''),
                        'module_modname': mod_name,
                    }
                )

                # Handle not supported modules that results in an index.html special
                if location['module_modname'] in ['page'] and self.version < 2017051500:
                    location['module_modname'] = 'index_mod-' + location['module_modname']

                files += self._handle_files(module.get('files', []), **location)

        return files

    @staticmethod
    def filter_changing_attributes(description: str) -> str:
        """
        Tries to filter ids and stuff,
        that is knowing to change over time in descriptions.
        """
        if description is None:
            return ''

        if not isinstance(description, str):
            return description

        # to avoid changing encodings (see issue #96) we unencode and unquote everything
        description = html.unescape(description)
        description = urlparse.unquote(description)

        # ids can change very quickly
        description = re.sub(r'id="[^"]*"', "", description)
        description = re.sub(r"id='[^']*'", "", description)

        # Embedded images from Moodle can change their timestemp (is such a theme feature)
        # We change every timestemp to -1 the default.
        description = re.sub(
            r"\/theme\/image.php\/(\w+)\/(\w+)\/\d+\/",
            r"/theme/image.php/\g<1>/\g<2>/-1/",
            description,
        )

        # some folder downloads inside a description file may have some session key inside which will always be
        # different. We remove it, to prevent always tagging this file as "modified".
        description = re.sub(r'<input type="hidden" name="sesskey" value="[0-9a-zA-Z]*" \/>', "", description)
        description = re.sub(r"<input type='hidden' name='sesskey' value='[0-9a-zA-Z]*' \/>", "", description)

        return description

    def _find_all_urls(
        self,
        content_html: str,
        no_search_for_moodle_urls: bool,
        filter_urls_containing: List[str],
        **location,
    ) -> List[File]:
        """Parses a html string to find all urls in it. Then it creates for every url a file entry.

        @param location: contains
            section_id: int,
            section_name: str,
            module_id: str,
            module_name: str,
            module_modname: str,
            content_filepath: str,
        """

        # TODO: Also parse name or alt of an link to get a better name for URLs
        urls = list(set(re.findall(r'href=[\'"]?([^\'" >]+)', content_html)))
        urls += list(set(re.findall(r'<a[^>]*>(http[^<]*)<\/a>', content_html)))
        urls += list(set(re.findall(r'src=[\'"]?([^\'" >]+)', content_html)))
        urls += list(set(re.findall(r'data=[\'"]?([^\'" >]+)', content_html)))
        urls = list(set(urls))

        result = []
        original_module_modname = location['module_modname']

        for url in urls:
            if url == '':
                continue

            # To avoid different encodings and quotes and so that yt-dlp downloads correctly
            # (See issues #96 and #103), we remove all encodings.
            url = html.unescape(url)
            url = urlparse.unquote(url)

            url_parts = urlparse.urlparse(url)
            if (
                url_parts.hostname == self.moodle_domain
                or url_parts.netloc == self.moodle_domain
                and no_search_for_moodle_urls
            ):
                # Skip if no moodle urls should be found
                continue

            for filter_str in filter_urls_containing:
                # Skip url if a filter matches
                if url.find(filter_str) >= 0:
                    continue

            if url_parts.hostname == self.moodle_domain and url_parts.path.find('/theme/image.php/') >= 0:
                url = re.sub(
                    r"\/theme\/image.php\/(\w+)\/(\w+)\/\d+\/",
                    r"/theme/image.php/\g<1>/\g<2>/-1/",
                    url,
                )

            location['module_modname'] = 'url-description-' + original_module_modname

            if url_parts.hostname == self.moodle_domain and url_parts.path.find('/webservice/') >= 0:
                location['module_modname'] = 'index_mod-description-' + original_module_modname

            elif url_parts.hostname == self.moodle_domain:
                location['module_modname'] = 'cookie_mod-description-' + original_module_modname

            if url.startswith('data:'):
                # Schema: data:[<mime type>][;charset=<Charset>][;base64],<Data>
                embedded_data = url.split(',', 1)[1]
                mime_type = url.split(':', 1)[1].split(',', 1)[0].split(';')[0]
                media_type = mime_type.split('/', 1)[0]
                file_extension_guess = mimetypes.guess_extension(mime_type, strict=False)
                if file_extension_guess is None:
                    file_extension_guess = f'.{media_type}'
                m = hashlib.sha1()
                if len(embedded_data) > 100000:
                    # To improve speed hash only first 100kb if file is bigger
                    m.update(embedded_data[:100000].encode(encoding='utf-8'))
                else:
                    m.update(embedded_data.encode(encoding='utf-8'))
                short_data_hash = m.hexdigest()

                fist_guess_filename = f'embedded_{media_type} ({short_data_hash}){file_extension_guess}'
            else:
                fist_guess_filename = url
                if len(fist_guess_filename) > 254:
                    fist_guess_filename = fist_guess_filename[:254]

            result.append(
                File(
                    **location,
                    content_filename=fist_guess_filename,
                    content_fileurl=url,
                    content_filesize=0,
                    content_timemodified=0,
                    content_type='description-url',
                    content_isexternalfile=True,
                )
            )
        return result

    def _handle_cookie_mod(self, module_url: str, **location) -> List[File]:
        """
        Create a list of files out of a cookie module
        @param module_url: The url to the cookie module.
        @param location: contains
            section_id: int,
            section_name: str,
            module_id: str,
            module_name: str,
            module_modname: str,
        """
        return [
            File(
                **location,
                content_filepath='/',
                content_filename=location['module_name'],
                content_fileurl=module_url,
                content_filesize=0,
                content_timemodified=0,
                content_type='cookie_mod',
                content_isexternalfile=True,
            )
        ]

    def _handle_files(self, module_contents: List, **location) -> List[File]:
        """
        Create a list of all files in a module
        @param module_contents: The list of content of the module
        @param location: contains
            section_id: int,
            section_name: str,
            module_id: str,
            module_name: str,
            module_modname: str,
        """
        files = []
        for content in module_contents:
            content_type = content.get('type', '')
            content_filename = content.get('filename', '')
            content_filepath = content.get('filepath', '/') or '/'
            content_fileurl = content.get('fileurl', '')

            content_description = content.get('description', '')
            content_html = content.get('html', '')

            if content_fileurl == '' and location['module_modname'].startswith(('url', 'index_mod', 'cookie_mod')):
                continue

            # Add the extention condition to avoid renaming pdf files or other downloaded content from moodle pages.
            if location['module_modname'].startswith('index_mod') and content_filename.endswith('.html'):
                content_filename = location['module_name']

            file_hash = None
            if content_type in ('description', 'html') and not content.get('no_hash', False):
                hashable_description = self.filter_changing_attributes(content_description)
                m = hashlib.sha1()
                m.update(hashable_description.encode('utf-8'))
                file_hash = m.hexdigest()

            new_file = File(
                **location,
                content_filepath=content_filepath,
                content_filename=content_filename,
                content_fileurl=content_fileurl,
                content_filesize=content.get('filesize', 0),
                content_timemodified=content.get('timemodified', 0),
                content_type=content_type,
                content_isexternalfile=content.get('isexternalfile', False),
                file_hash=file_hash,
            )

            if content_type == 'description':
                new_file.text_content = content_description
                content_html = content_description
            if content_type == 'html':
                new_file.html_content = content_html

            if content_type in ['description', 'html'] and not content.get('no_search_for_urls', False):
                files += self._find_all_urls(
                    content_html,
                    no_search_for_moodle_urls=content.get('no_search_for_moodle_urls', False),
                    filter_urls_containing=content.get('filter_urls_during_search_containing', []),
                    **location,
                    content_filepath=content_filepath,
                )

            files.append(new_file)
        return files

    def _handle_description(
        self,
        module_description: str,
        **location,
    ) -> List[File]:
        """
        Creates a description file
        @param module_description: The description of the module
        @param location: contains
            section_id: int,
            section_name: str,
            module_id: str,
            module_name: str,
            module_modname: str,
        @return: A list of files containing that description and URLs in that description.
        """
        files = []
        content_filepath = '/'

        m = hashlib.sha1()
        hashable_description = self.filter_changing_attributes(module_description)
        m.update(hashable_description.encode('utf-8'))
        hash_description = m.hexdigest()

        if location['module_modname'].startswith(('url', 'index_mod')):
            location['module_modname'] = 'url_description'

        description = File(
            **location,
            content_filepath=content_filepath,
            content_filename=location['module_name'],
            content_fileurl='',
            content_filesize=len(module_description),
            content_timemodified=0,
            content_type='description',
            content_isexternalfile=False,
            file_hash=hash_description,
        )
        description.text_content = module_description
        files.append(description)

        files += self._find_all_urls(
            module_description,
            no_search_for_moodle_urls=False,
            filter_urls_containing=[],
            **location,
            content_filepath=content_filepath,
        )

        return files

    def add_files_to_courses(
        self,
        courses: List[Course],
        course_cores: Dict[int, List[Dict]],
        fetched_mods_files: Dict[str, Dict],
    ):
        """
        @param fetched_mods_files:
            Dictionary of all fetched mod modules files, indexed by mod name, then by courses, then module id
        """
        for course in courses:
            course_sections = course_cores.get(course.id, [])

            fetched_mods = {}
            for mod_name, mod_courses in fetched_mods_files.items():
                fetched_mods[mod_name] = mod_courses.get(course.id, {})

            course.files = self.get_files_in_sections(course_sections, fetched_mods)



import logging
import os
from typing import Dict

from moodle_dl.config import ConfigHelper
from moodle_dl.moodle.request_helper import RequestHelper, RequestRejectedError
from moodle_dl.types import MoodleDlOpts
from moodle_dl.utils import PathTools as PT


class CookieHandler:
    """
    Fetches and saves the cookies of Moodle.
    """

    def __init__(self, request_helper: RequestHelper, version: int, config: ConfigHelper, opts: MoodleDlOpts):
        self.client = request_helper
        self.version = version
        self.config = config
        self.opts = opts
        self.cookies_path = PT.get_cookies_path(config.get_misc_files_path())

        self.moodle_test_url = self.client.url_base

    def fetch_autologin_key(self, privatetoken: str) -> Dict[str, str]:
        if self.version < 2016120500:  # 3.2
            return None

        logging.info('Downloading autologin key')

        extra_data = {'privatetoken': privatetoken}

        try:
            autologin_key_result = self.client.post('tool_mobile_get_autologin_key', extra_data)
            return autologin_key_result
        except RequestRejectedError as e:
            logging.debug("Cookie lockout: %s", e)
            return None

    def test_cookies(self) -> bool:
        """
        Test if cookies are valid
        @return: True if valid
        """

        logging.debug('Testing cookies using this URL: %s', self.moodle_test_url)

        response, dummy = self.client.get_URL(self.moodle_test_url, self.cookies_path)

        response_text = response.text

        if response_text.find('login/logout.php') >= 0:
            return True
        return False

    def check_and_fetch_cookies(self, privatetoken: str, userid: str) -> bool:
        if os.path.exists(self.cookies_path):
            if self.test_cookies():
                logging.debug('Cookies are still valid')
                return True

            logging.info('Moodle cookie has expired, an attempt is made to generate a new cookie.')

        if privatetoken is None:
            error_msg = (
                'Moodle Cookies are not retrieved because no private token is set.'
                + ' To set a private token, use the `--new-token` option (if necessary also with `--sso`).'
            )
            logging.debug(error_msg)
            return False

        autologin_key = self.fetch_autologin_key(privatetoken)

        if autologin_key is None:
            logging.debug('Failed to download autologin key!')
            return False

        logging.info('Downloading cookies')

        post_data = {'key': autologin_key.get('key', ''), 'userid': userid}
        url = autologin_key.get('autologinurl', '')

        cookies_response, _ = self.client.post_URL(url, post_data, self.cookies_path)

        logging.debug('Autologin redirected to %s', cookies_response.url)

        if self.test_cookies():
            return True

        logging.debug('Failed to generate cookies!')
        return False



from typing import List

from moodle_dl.notifications.telegram.telegram_formater import TelegramFormater


class XmppFormater(TelegramFormater):
    @staticmethod
    def append_with_limit(new_line: str, one_msg_content: str, msg_list: List[str], limit: int = 4096):
        """Appends a new line to a message string,
        if the string is to long it ist appended to the message list.
        Returns the new message string.

        Args:
            new_line (str): A new line to append to a message string
            one_msg_content (str): The current message string
            msg_list ([str]): The list of finished messages
        Returns:
            str: The new message string
        """
        if len(one_msg_content) + len(new_line) >= limit:
            msg_list.append(one_msg_content)
            if len(new_line) >= limit:
                new_line = new_line[: limit - 3] + '…'
            return new_line
        else:
            return one_msg_content + new_line

    @staticmethod
    def make_bold(string: str) -> str:
        """
        Makes a string bold in a xmpp message
        """
        return '*' + string + '*'



import os

from moodle_dl.config import ConfigHelper
from moodle_dl.database import StateRecorder
from moodle_dl.moodle.moodle_service import MoodleService
from moodle_dl.types import File, MoodleDlOpts
from moodle_dl.utils import Cutie, Log


class DatabaseManager:
    def __init__(self, config: ConfigHelper, opts: MoodleDlOpts):
        self.config = config
        self.opts = opts
        self.state_recorder = StateRecorder(config, opts)

    def interactively_manage_database(self):
        stored_files = self.state_recorder.get_stored_files()

        stored_files = MoodleService.filter_courses(stored_files, self.config)

        if len(stored_files) <= 0:
            return

        course_options = []
        courses = []
        for course in stored_files:
            for course_file in course.files:
                if not os.path.exists(course_file.saved_to):
                    course_options.append(Log.blue_str(course.fullname))
                    courses.append(course)
                    break

        print(
            'This management tool will navigate you through a menu to'
            + ' selectively remove file entries from the database so'
            + ' that these files can be downloaded again.'
        )

        Log.warning(
            'Only files that are missing locally but stored in the local'
            + ' database are displayed in this tool. If a file is not missing'
            + ' from a course, it will not be listed here at all.  Also, only'
            + ' courses that are selected for download are displayed.'
        )

        Log.magenta(
            'For more complicated operations on the database a DB browser for SQLite'
            + ' is advantageous (https://sqlitebrowser.org/).'
        )

        if not courses:
            print('No files are missing locally but stored in the local database. Nothing to do.')
            return

        print('Choose one of the courses:')
        print('[Confirm your selection with the Enter key]')
        print('')
        selected_course_id = Cutie.select(options=course_options)

        selected_course = courses[selected_course_id]

        section_options = []
        sections = []

        # Add the option to select all sections
        section_options.append(Log.magenta_str('[All sections]'))
        sections.append(None)  # Add None at index 0 to avoid index shifting

        for course_file in selected_course.files:
            if not os.path.exists(course_file.saved_to) and (course_file.section_name not in sections):
                section_options.append(Log.magenta_str(course_file.section_name))
                sections.append(course_file.section_name)

        print('From which sections you want to select files?')
        print('[You can select with the space bar and confirm your selection with the enter key]')
        print('')

        selected_sections_ids = Cutie.select_multiple(options=section_options, minimal_count=1)

        selected_sections = []
        for selected_sections_id in selected_sections_ids:
            if selected_sections_id == 0:
                selected_sections = sections[1:]
                break
            elif (selected_sections_id) < len(sections):
                selected_sections.append(sections[selected_sections_id])

        file_options = []
        files = []

        # Add the option to select all files
        file_options.append(Log.cyan_str('[All files]'))
        files.append(None)  # Add None at index 0 to avoid index shifting

        for course_file in selected_course.files:
            if not os.path.exists(course_file.saved_to) and (course_file.section_name in selected_sections):
                file_options.append(Log.cyan_str(course_file.content_filename))
                files.append(course_file)

        print('Which of the files should be removed form the database, so that they will be re-downloaded?')
        print('[You can select with the space bar and confirm your selection with the enter key]')
        print('')
        selected_files = Cutie.select_multiple(options=file_options)

        files_to_delete = []
        for file_index in selected_files:
            if file_index == 0:  # If all files is selected
                for file_to_delete in files[1:]:  # Ignore the first element of the array set as None
                    if isinstance(file_to_delete, File):
                        files_to_delete.append(file_to_delete)

                break

            elif file_index < len(files) and isinstance(files[file_index], File):
                files_to_delete.append(files[file_index])

        self.state_recorder.batch_delete_files_from_db(files_to_delete)

    def delete_old_files(self):
        stored_files = self.state_recorder.get_old_files()

        if len(stored_files) <= 0:
            print('No old copies of files found. Nothing to do.')
            return

        print(
            'This management tool will navigate you through a menu to'
            + ' selectively remove old copies of files from the database '
            + ' and form the file system'
        )

        course_options = []
        for course in stored_files:
            course_options.append(Log.blue_str(course.fullname))

        print('Choose one of the courses:')
        print('[Confirm your selection with the Enter key]')
        print('')
        selected_course_id = Cutie.select(options=course_options)

        selected_course = stored_files[selected_course_id]

        section_options = []
        sections = []

        # Add the option to select all sections
        section_options.append(Log.magenta_str('[All sections]'))
        sections.append(None)  # Add None at index 0 to avoid index shifting

        for course_file in selected_course.files:
            if course_file.section_name not in sections:
                section_options.append(Log.magenta_str(course_file.section_name))
                sections.append(course_file.section_name)

        print('From which sections you want to delete old files?')
        print('[You can select with the space bar and confirm your selection with the enter key]')
        print('')

        selected_sections_ids = Cutie.select_multiple(options=section_options, minimal_count=1)

        selected_sections = []
        for selected_sections_id in selected_sections_ids:
            if selected_sections_id == 0:
                selected_sections = sections[1:]
                break
            elif (selected_sections_id) < len(sections):
                selected_sections.append(sections[selected_sections_id])

        file_options = []
        files = []

        # Add the option to select all files
        file_options.append(Log.cyan_str('[All files]'))
        files.append(None)  # Add None at index 0 to avoid index shifting

        for course_file in selected_course.files:
            if course_file.section_name in selected_sections:
                file_options.append(Log.cyan_str(course_file.content_filename))
                files.append(course_file)

        print('Which of the files should be deleted?')
        print('[You can select with the space bar and confirm your selection with the enter key]')
        print('')
        selected_files = Cutie.select_multiple(options=file_options)

        files_to_delete = []
        for file_index in selected_files:
            if file_index == 0:  # If all files is selected
                for file_to_delete in files[1:]:  # Ignore the first element of the array set as None
                    if isinstance(file_to_delete, File):
                        files_to_delete.append(file_to_delete)
                        if os.path.exists(file_to_delete.saved_to):
                            os.remove(file_to_delete.saved_to)

                break

            elif file_index < len(files) and isinstance(files[file_index], File):
                files_to_delete.append(files[file_index])
                if os.path.exists(files[file_index].saved_to):
                    os.remove(files[file_index].saved_to)

        self.state_recorder.batch_delete_files_from_db(files_to_delete)



import json
import os
import sys
from pathlib import Path
from typing import Dict, List, Tuple

from moodle_dl.types import DownloadOptions, MoodleDlOpts, MoodleURL
from moodle_dl.utils import PathTools as PT


class ConfigHelper:
    "Handles the saving, formatting and loading of the local configuration."

    class NoConfigError(ValueError):
        """An Exception which gets thrown if config could not be loaded."""

        pass

    def __init__(self, opts: MoodleDlOpts):
        self._whole_config = {}
        self.opts = opts
        self.config_path = str(Path(opts.path) / 'config.json')

    def is_present(self) -> bool:
        # Tests if a configuration file exists
        return os.path.isfile(self.config_path)

    def load(self):
        # TODO: Load config into dataclass, so we can access that class instead of using getters
        # Opens the configuration file and parse it to a JSON object
        try:
            with open(self.config_path, 'r', encoding='utf-8') as config_file:
                config_raw = config_file.read()
                self._whole_config = json.loads(config_raw)
        except (IOError, OSError) as err_load:
            raise ConfigHelper.NoConfigError(f'Configuration could not be loaded from {self.config_path}\n{err_load!s}')

    def _save(self):
        # TODO: Use dataclass and write that back to file, so that all options are always present
        config_formatted = json.dumps(self._whole_config, indent=4)
        # Saves the JSON object back to file
        with os.fdopen(
            os.open(self.config_path, flags=os.O_WRONLY | os.O_CREAT, mode=0o600), mode='w', encoding='utf-8'
        ) as config_file:
            config_file.write(config_formatted)

    def get_property(self, key: str) -> any:
        # return a property if configured
        try:
            return self._whole_config[key]
        except KeyError:
            raise ValueError(f'The {key}-Property is not yet configured!')

    def get_property_or(self, key: str, default: any = None) -> any:
        # return a property if configured
        try:
            return self._whole_config[key]
        except KeyError:
            return default

    def set_property(self, key: str, value: any):
        # sets a property in the JSON object
        self._whole_config.update({key: value})
        self._save()

    def remove_property(self, key):
        # removes a property from the JSON object
        self._whole_config.pop(key, None)
        #                           ^ behavior if the key is not present
        self._save()

    # ---------------------------- GETTERS ------------------------------------

    def get_download_submissions(self) -> str:
        # return a stored boolean if submissions should be downloaded
        return self.get_property_or('download_submissions', False)

    def get_download_descriptions(self) -> bool:
        # return a stored boolean if descriptions should be downloaded
        return self.get_property_or('download_descriptions', False)

    def get_download_links_in_descriptions(self) -> bool:
        # return a stored boolean if links in descriptions should be downloaded
        return self.get_property_or('download_links_in_descriptions', False)

    def get_download_databases(self) -> bool:
        # return a stored boolean if databases should be downloaded
        return self.get_property_or('download_databases', False)

    def get_download_forums(self) -> bool:
        # return a stored boolean if forums should be downloaded
        return self.get_property_or('download_forums', False)

    def get_download_quizzes(self) -> bool:
        # return a stored boolean if quizzes should be downloaded
        return self.get_property_or('download_quizzes', False)

    def get_download_lessons(self) -> bool:
        # return a stored boolean if lessons should be downloaded
        return self.get_property_or('download_lessons', False)

    def get_download_workshops(self) -> bool:
        # return a stored boolean if workshops should be downloaded
        return self.get_property_or('download_workshops', False)

    def get_userid_and_version(self) -> Tuple[str, int]:
        # return the userid and a version
        try:
            user_id = self.get_property('userid')
            version = int(self.get_property('version'))
            return user_id, version
        except ValueError:
            return None, None

    def get_do_not_ask_to_save_userid_and_version(self) -> bool:
        return self.get_property_or('do_not_ask_to_save_userid_and_version', False)

    def get_download_course_ids(self) -> str:
        # return a stored list of course ids hat should be downloaded
        return self.get_property_or('download_course_ids', [])

    def get_download_public_course_ids(self) -> str:
        # return a stored list of public course ids hat should be downloaded
        return self.get_property_or('download_public_course_ids', [])

    def get_token(self) -> str:
        # return a stored token
        try:
            return self.get_property('token')
        except ValueError:
            raise ValueError('Not yet configured!')

    def get_privatetoken(self) -> str:
        # return a stored privatetoken
        return self.get_property_or('privatetoken', None)

    def get_moodle_URL(self) -> MoodleURL:
        moodle_domain = self.get_moodle_domain()
        moodle_path = self.get_moodle_path()
        use_http = self.get_use_http()
        return MoodleURL(use_http, moodle_domain, moodle_path)

    def get_moodle_domain(self) -> str:
        # return a stored moodle_domain
        try:
            return self.get_property('moodle_domain')
        except ValueError:
            raise ValueError('Not yet configured!')

    def get_moodle_path(self) -> str:
        # return a stored moodle_path
        try:
            return self.get_property('moodle_path')
        except ValueError:
            raise ValueError('Not yet configured!')

    def get_options_of_courses(self) -> Dict:
        # return a stored dictionary of options for courses
        return self.get_property_or('options_of_courses', {})

    def get_dont_download_course_ids(self) -> List:
        # return a stored list of ids that should not be downloaded
        return self.get_property_or('dont_download_course_ids', [])

    def get_download_linked_files(self) -> bool:
        # return if linked files should be downloaded
        return self.get_property_or('download_linked_files', False)

    def get_download_domains_whitelist(self) -> List:
        # return a list of white listed domains that should be downloaded
        return self.get_property_or('download_domains_whitelist', [])

    def get_download_domains_blacklist(self) -> List:
        # return a list of black listed domains that should not be downloaded
        return self.get_property_or('download_domains_blacklist', [])

    def get_cookies_text(self) -> str:
        # return the text to the cookies file, if it exists
        cookies_path = PT.get_cookies_path(self.get_misc_files_path())
        if os.path.isfile(cookies_path):
            with open(cookies_path, 'r', encoding='utf-8') as cookie_file:
                return cookie_file.read()
        return None

    def get_yt_dlp_options(self) -> Dict:
        # return additional yt-dlp options
        return self.get_property_or('yt_dlp_options', {})

    def get_video_passwords(self) -> Dict:
        # return dict with passwords that get passed to yt-dlp
        return self.get_property_or('video_passwords', {})

    def get_external_file_downloaders(self) -> Dict:
        # return dict with configured external downloaders
        return self.get_property_or('external_file_downloaders', {})

    def get_exclude_file_extensions(self) -> Dict:
        # return a list of file extensions that should not be downloaded
        try:
            exclude_file_extensions = self.get_property('exclude_file_extensions')
            if not isinstance(exclude_file_extensions, list):
                exclude_file_extensions = [exclude_file_extensions]
            return exclude_file_extensions
        except ValueError:
            return []

    def get_max_file_size(self) -> int:
        # return the max size in bytes of files that should not be downloaded
        # default: 0 -> all file sizes
        return self.get_property_or('max_file_size', 0)

    def get_download_also_with_cookie(self) -> Dict:
        # return if files for which a cookie is required should be downloaded
        return self.get_property_or('download_also_with_cookie', False)

    def get_write_links(self) -> Dict:
        # returns what kind of shortcuts should be created
        write_links = {
            'url': self.get_property_or('write_url_link', False),
            'webloc': self.get_property_or('write_webloc_link', False),
            'desktop': self.get_property_or('write_desktop_link', False),
        }
        if self.get_property_or('write_link', True):
            link_type = (
                'webloc' if sys.platform == 'darwin' else 'desktop' if sys.platform.startswith('linux') else 'url'
            )
            write_links[link_type] = True

        return write_links

    def get_download_options(self, opts: MoodleDlOpts) -> DownloadOptions:
        # return the option dictionary for downloading files

        return DownloadOptions(
            token=self.get_token(),
            download_linked_files=self.get_download_linked_files(),
            download_domains_whitelist=self.get_download_domains_whitelist(),
            download_domains_blacklist=self.get_download_domains_blacklist(),
            cookies_text=self.get_cookies_text(),
            yt_dlp_options=self.get_yt_dlp_options(),
            video_passwords=self.get_video_passwords(),
            external_file_downloaders=self.get_external_file_downloaders(),
            restricted_filenames=self.get_restricted_filenames(),
            write_links=self.get_write_links(),
            download_path=self.get_download_path(),
            global_opts=opts,
        )

    def get_restricted_filenames(self) -> Dict:
        # return the filenames should be restricted
        return self.get_property_or('restricted_filenames', False)

    def get_use_http(self) -> bool:
        # return a stored boolean if http should be used instead of https
        return self.get_property_or('use_http', False)

    def get_download_path(self) -> str:
        # return path of download location
        return self.get_property_or('download_path', self.opts.path)

    def get_misc_files_path(self) -> str:
        # return path of misc files
        return self.get_property_or('misc_files_path', self.opts.path)

    # ---------------------------- SETTERS ------------------------------------

    def set_moodle_URL(self, moodle_url: MoodleURL):
        self.set_property('moodle_domain', moodle_url.domain)
        self.set_property('moodle_path', moodle_url.path)
        if moodle_url.use_http is True:
            self.set_property('use_http', moodle_url.use_http)
        else:
            if self.get_use_http():
                self.set_property('use_http', moodle_url.use_http)

    def set_tokens(self, moodle_token: str, moodle_privatetoken: str):
        self.set_property('token', moodle_token)
        if moodle_privatetoken is not None:
            self.set_property('privatetoken', moodle_privatetoken)



from typing import List

from moodle_dl.config import ConfigHelper
from moodle_dl.notifications.console.console_service import ConsoleService
from moodle_dl.notifications.discord.discord_service import DiscordService
from moodle_dl.notifications.mail.mail_service import MailService
from moodle_dl.notifications.notification_service import NotificationService
from moodle_dl.notifications.telegram.telegram_service import TelegramService
from moodle_dl.notifications.xmpp.xmpp_service import XmppService

__all__ = ['ConsoleService', 'MailService', 'TelegramService', 'DiscordService', 'XmppService']

REMOTE_SERVICES = [
    Class
    for name, Class in globals().items()
    if name.endswith('Service') and name not in ['ConsoleService', 'NotificationService']
]
ALL_SERVICES = [ConsoleService] + REMOTE_SERVICES


def get_remote_notify_services(config: ConfigHelper) -> List[NotificationService]:
    result_list = []
    for service in REMOTE_SERVICES:
        result_list.append(service(config))
    return result_list


def get_all_notify_services(config: ConfigHelper) -> List[NotificationService]:
    result_list = []
    for service in ALL_SERVICES:
        result_list.append(service(config))
    return result_list
